﻿using System.Collections.Generic;
using Stride.Core.Mathematics;
using VL.Lib.Mathematics;


namespace ForceDirected
{
    public class Octree
    {
        private Octree[] FChildren = null;  // Represents the subtrees. If null then this node is a leaf

        private BoundingBox FBoundary;      // bounds in which a body can be placed in
        private float FSize;

        private int FDepth;                 // How deep this subtree is in the whole tree

        private FDNode FBody;               // Body represented in this node

        private float FTheta = 0.8f;

        private Vector3 FCenterOfMass;
        private float FTotalMass;

        private int FBodyCount;


        public Octree[] Children
        {
            get { return FChildren; }
        }


        public BoundingBox Boundary
        {
            get { return FBoundary; }
        }


        public int Depth
        {
            get { return FDepth; }
            set { FDepth = value; }
        }


        public Vector3 CenterOfMass
        {
            get { return FCenterOfMass; }
        }


        public float TotalMass
        {
            get { return FTotalMass; }
        }


        public int NodeCount
        {
            get { return FBodyCount; }
        }

        public float Theta
        {
            get { return FTheta; }
        }


        public float Size
        {
            get { return FSize; }
        }


        private Octree(Vector3 center, Vector3 size, float theta)
        {
            FBoundary = AlignedBoxExtensions.CreateCenterSize(ref center, ref size);
            FTheta = theta;
            SetSize();
        }


        public Octree(Vector3 min, Vector3 max, int depth, float theta)
        {
            FBoundary = new BoundingBox(min, max);
            FDepth = depth;
            FTheta = theta;
            SetSize();
        }


        public Octree(Vector3[] positions, int depth, float theta)
        {
            BoundingBox.FromPoints(positions, out FBoundary);
            FDepth = depth;
            FTheta = theta;
            SetSize();
        }

        
        private void SetSize()
        {
            // used to calulate theta
            FSize = Vector3.DistanceSquared(FBoundary.Minimum, FBoundary.Maximum);
        }

        private List<FDNode> ContainsNodes(List<FDNode> nodes)
        {
            List<FDNode> containedNodes = new List<FDNode>();

            foreach (FDNode node in nodes)
            {
                if (FBoundary.Contains(ref node.Position) == ContainmentType.Contains)
                {
                    FCenterOfMass = (FTotalMass * FCenterOfMass + node.Mass * node.Position) / (FTotalMass + node.Mass);
                    FTotalMass += node.Mass;

                    FBodyCount++;

                    containedNodes.Add(node);
                }
            }

            return containedNodes;
        }

        // Subdivide BoundingBox into 8 octants
        private void Subdivide()
        {
            Vector3[] corners = FBoundary.GetCorners();
            Vector3 boundaryCenter = FBoundary.Center;
            Vector3 boundaryMaximum = FBoundary.Maximum;

            Vector3 size = boundaryMaximum - boundaryCenter;

            FChildren = new Octree[8];

            for (int i = 0; i < 8; i++)
            {
                Vector3 center = (corners[i] + boundaryCenter) * 0.5f;

                Octree child = new Octree(center, size, FTheta);
                child.Depth = FDepth + 1;

                FChildren[i] = child;
            }
        }


        public void BuildTree(List<FDNode> bodies)
        {
            List<FDNode> containedBodies = ContainsNodes(bodies);

            if(FBodyCount == 1)
            {
                FBody = containedBodies[0];    
            }

            if(FBodyCount > 1)
            { 
                Subdivide();

                foreach (Octree branch in FChildren) 
                {
                    branch.BuildTree(containedBodies);
                }
            }
        }

        public Vector3 CalcRepulsion(FDNode body)
        {
            Vector3 repulsion = new Vector3(0);

            // make root node stay in place || or octant doesn't contain any bodies
            if (body.Parent.Count == 0 || FBodyCount == 0)
                return repulsion;


            Vector3 position = body.Position;
            
           
            if (FBodyCount == 1)
            {
                if (body == FBody)
                    return repulsion;

                Vector3 diff = position - FBody.Position;

                return repulsion += Forces.Repulse(diff);
            }
            else 
            {
                Vector3 diff = position - FCenterOfMass;
                float r = diff.LengthSquared();
                float d = FSize;

                if(d/r < FTheta)
                {
                    return repulsion += Forces.Repulse(diff) * FTotalMass;
                }
                else
                {
                    foreach (Octree branch in FChildren)
                    {
                        repulsion += branch.CalcRepulsion(body);
                    }
                    return repulsion;
                }
            }

        }
    }
}
